-----------------------------------------------------------------------------------------
{-| Module      :  Controls
    Copyright   :  (c) Krasimir Angelov 2003
    License     :  BSD-style

    Maintainer  :  ka2_mail@yahoo.com
    Stability   :  provisional
    Portability :  portable

    This module contains interface to all controls supported from GIO.
    A control is a child window an application uses in conjunction with
    another window to perform simple input and output tasks.
    The controls provides the user with the means to type text, choose options,
    and execute an actions.
-}
-----------------------------------------------------------------------------------------
module Graphics.UI.GIO.Controls(
                               -- * Controls
                               -- ** Button
                               -- | A button control is a small, rectangular child window that can be
                               -- clicked on and off. Buttons can be labeled or appear without text.
                               -- A button typically changes appearance when the user clicks it.
                                 Button, button

                               -- ** Label
                               -- | A label control simply displays a text string and it can be used
                               -- to label other controls. A label control can\'t take input and can\'t provide output.
                               , Label, label

                               -- ** Entry
                               -- | An entry control is a rectangular widget in which the user can enter text.
                               , Entry, entry, readOnly, password

                               -- ** Popup
                               -- | A popup control consists of a list box combined with a label control.
                               -- The list box portion of the control drop down only when the user selects 
                               -- the drop-down arrow next to the control. The currently selected item in the 
                               -- list box is displayed in the label control.
                               , Popup, popup
                               
                               -- ** ListBox
                               -- | A list box displays a list of items, that the user can view and select.
                               -- In a single-selection list box, the user can select only one item. 
                               -- In a multiple-selection list box, a range of items can be selected. 
                               -- When the user selects an item, it is highlighted and the list box 
                               -- the 'command' event is generated.
                               , ListBox, listBox
                               
                               -- ** Slider
                               -- | A slider control is a widget containing a slider and tick marks.
                               -- When the user moves the slider, using either the mouse or the direction keys,
                               -- the control generates the 'command' event to indicate the change.
                               -- Slider controls are useful when you want the user to select a discrete value 
                               -- in a given range. The slider can be vertical or horizontal.
                               , Slider, hslider, vslider
                               
                               -- ** ProgressBar
                               -- | A progress bar control is a window that an application can use to 
                               -- indicate the progress of a lengthy operation. It consists of a rectangle 
                               -- that is gradually filled, from left to right, with the system specified color
                               -- as an operation progresses. A progress bar control has a range and a current 
                               -- position. The range represents the entire duration of the operation, and the 
                               -- current position represents the progress the application has made toward 
                               -- completing the operation.
                               , ProgressBar, hProgressBar, vProgressBar
                               
                               -- ** CheckBox
                               -- | A check box consists of a square box and a label, that indicates a choice 
                               -- the user can make by selecting the check. Each CheckBox has two states: 
                               -- checked (a check mark inside the box) or cleared (no check mark). The
                               -- state is controled from the 'checked' attribute. Repeatedly clicking a 
                               -- check box toggles it from checked to cleared and back again.
                               , CheckBox, checkBox
                               
                               -- ** RadioBox
                               -- | A radio button consists of a round button and a label that indicates a 
                               -- choice the user can make by selecting the button. An application typically 
                               -- uses radio buttons in a group box to permit the user to choose from a set 
                               -- of mutually exclusive options. The group is defined with 'setRadioBoxGroup'
                               -- function. When the user selects an radio button from group then the system 
                               -- automatically sets its check state to 'checked' and clears all other 
                               -- buttons within the same group.
                               , RadioBox, radioBox, setRadioBoxGroup
                               
                               -- ** GroupBox
                               -- | A group box is a rectangle that surrounds a set of controls, such as 
                               -- check boxes or radio buttons, with application-defined label in its upper 
                               -- left corner. The sole purpose of a group box is to organize controls related 
                               -- by a common purpose (usually indicated by the label).
                               , GroupBox, groupBox
                               
                               -- ** CompoundControl
                               , CompoundControl, compoundControl
                               
                               -- ** Notebook
                               -- | A notebook control is analogous to the dividers in a real notebook.
                               -- By using a notebook control, an application can define multiple pages for 
                               -- the same area of a window or dialog box. Each page is represented from 
                               -- 'NotebookPage' type and is created from 'notebookPageAt' or 'notebookPage'
                               -- function.
                               , Notebook, notebook, labelsPosition, selectedPage
                               , NotebookPage, notebookPageAt, notebookPage
                               ) where

import qualified Graphics.UI.Port as Port
import Graphics.UI.GIO.Types
import Graphics.UI.GIO.Attributes
import Graphics.UI.GIO.Events
import Graphics.UI.GIO.Canvas
import Graphics.UI.GIO.Layout
import Control.Monad(when, mapM_, filterM)

--------------------------------------------------------------------
--  Helper functions
--------------------------------------------------------------------

insertAt :: Int -> a -> [a] -> [a]
insertAt pos v [] = []
insertAt pos v (x:xs)
  | pos == 0  = v : xs
  | otherwise = x : insertAt (pos-1) v xs
        
removeAt :: Int -> [a] -> [a]
removeAt pos [] = []
removeAt pos (x:xs)
  | pos == 0  = xs
  | otherwise = x : removeAt (pos-1) xs
      	  
--------------------------------------------------------------------
--  Label
--------------------------------------------------------------------
-- | A simple text label.
data Label    = Label { lhandle :: !WindowHandle
                      , lparent :: !WindowHandle
                      , lfont   :: Var Font
                      }

-- | Create a text label.
label :: Container w => [Prop Label] -> w -> IO Label
label props w
  = do lab <- do hlab <- Port.createLabel (hwindow w)
                 lfont <- newVar Port.defaultFont
                 return (Label hlab (hwindow w) lfont)
       set lab props
       return lab

instance Titled Label where
  title  = newStdAttr lhandle Port.getLabelText Port.setLabelText

instance HasFont Label where
  font  = newAttr (getVar . lfont) (\w font -> Port.changeLabelFont (lhandle w) font >> setVar (lfont w) font)

instance Dimensions Label where
  frame = newStdAttr lhandle Port.getControlFrame Port.moveResizeControl
                  
instance Able Label where
  enabled = newStdAttr lhandle Port.getControlEnabled Port.setControlEnabled
  
instance ToolTip Label where
  tooltip = newStdAttr lhandle Port.getControlTip Port.setControlTip

instance Control Label where
  pack = stdPack lparent lhandle Port.getLabelRequestSize

--------------------------------------------------------------------
--  Button
--------------------------------------------------------------------
-- | A standard push button.
data Button   = Button { bhandle :: !WindowHandle
                       , bparent :: !WindowHandle
                       , bfont   :: Var Font
                       }

-- | Create a button.
button :: Container w => [Prop Button] -> w -> IO Button
button props w
  = do but <- do hbut <- Port.createButton (hwindow w)
                 bfont <- newVar Port.defaultFont
                 return (Button hbut (hwindow w) bfont)
       set but props
       return but

instance Titled Button where
  title = newStdAttr bhandle Port.getButtonText Port.setButtonText

instance HasFont Button where
  font  = newAttr (getVar . bfont) (\w font -> Port.changeButtonFont (bhandle w) font >> setVar (bfont w) font)

instance Dimensions Button where
  frame = newStdAttr bhandle Port.getControlFrame Port.moveResizeControl

instance Able Button where
  enabled = newStdAttr bhandle Port.getControlEnabled Port.setControlEnabled
  
instance ToolTip Button where
  tooltip = newStdAttr bhandle Port.getControlTip Port.setControlTip

instance Control Button where
  pack = stdPack bparent bhandle Port.getButtonRequestSize
             
instance Commanding Button where
  command = newStdEvent bhandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler

--------------------------------------------------------------------
--  Entry
--------------------------------------------------------------------
-- | A standard text entry control.
data Entry   = Entry { ehandle :: !WindowHandle
                     , eparent :: !WindowHandle
                     , efont   :: Var Font
                     }

-- | Create an entry.
entry :: Container w => [Prop Entry] -> w -> IO Entry
entry props w
  = do e <- do hentry <- Port.createEdit (hwindow w)
               efont  <- newVar Port.defaultFont
               return (Entry hentry (hwindow w) efont)
       set e props
       return e

instance Titled Entry where
  title  = newStdAttr ehandle Port.getEditText Port.setEditText

instance HasFont Entry where
  font  = newAttr (getVar . efont) (\w font -> Port.changeEditBoxFont (ehandle w) font >> setVar (efont w) font)

instance Dimensions Entry where
  frame = newStdAttr ehandle Port.getControlFrame Port.moveResizeControl

instance Able Entry where
  enabled = newStdAttr ehandle Port.getControlEnabled Port.setControlEnabled
  
instance ToolTip Entry where
  tooltip = newStdAttr ehandle Port.getControlTip Port.setControlTip

instance Control Entry where
  pack = stdPack eparent ehandle Port.getEditRequestSize
    
-- | Determines if the user can edit the text in the editable widget or not. 
readOnly :: Attr Entry Bool    
readOnly = newStdAttr ehandle Port.getEditReadOnly Port.setEditReadOnly

-- | Determines whether the entry is used to edit a password or a normal text.
password :: Attr Entry Bool
password = newStdAttr ehandle Port.getEditPassword Port.setEditPassword

{--------------------------------------------------------------------
  Popup box
--------------------------------------------------------------------}
-- | A popup selection box. Allthough it is 'Commanding', the default
-- command handler automatically calls a handler associated with a
-- selected item.
data Popup  = Popup    { phandle  :: !WindowHandle
                       , pparent  :: !WindowHandle
                       , pitems   :: Var [(String,IO ())]
                       }

-- | Create a popup selection box.
popup :: Container w => [Prop Popup] -> w -> IO Popup
popup props w
  = do p <- do hpop <- Port.createPopUp (hwindow w)
               pitems <- newVar []
               return (Popup hpop (hwindow w) pitems)
       set p [on command =: popupCommand p]
       set p props
       return p

-- default command handler
popupCommand :: Popup -> IO ()
popupCommand p
  = do i  <- get p selected
       xs <- getVar (pitems p)
       when (i>=0 && i < length xs) (snd (xs!!i))  -- invoke appropiate handler

instance Countable Popup where
  count = readAttr "count" (fmap length . getVar . pitems)

instance CommandItems Popup where
  items
    = newAttr (\w -> getVar (pitems w))
              (\w xs -> do Port.removeAllPopUpItems (phandle w)
                           mapM_ (Port.appendPopUpItem (phandle w) . fst) xs
                           setVar (pitems w) xs
                           set w [selected =: 0])

  appendItem p item@(title,action) = do
    items <- takeVar (pitems p)
    Port.appendPopUpItem (phandle p) title
    putVar (pitems p) (items++[item])

  insertItem p pos item@(title,action) = do
    items <- takeVar (pitems p)
    Port.insertPopUpItem (phandle p) pos title
    putVar (pitems p) (insertAt pos item items)

  removeItem p pos = do
    items <- takeVar (pitems p)
    Port.removePopUpItem (phandle p) pos
    putVar (pitems p) (removeAt pos items)

  removeAllItems p = do
    items <- takeVar (pitems p)
    Port.removeAllPopUpItems (phandle p)
    putVar (pitems p) []

instance SingleSelect Popup where
  selected = newStdAttr phandle Port.getPopUpSelection Port.setPopUpSelection

instance Dimensions Popup where
  frame = newStdAttr phandle Port.getControlFrame Port.moveResizeControl

instance Able Popup where
  enabled = newStdAttr phandle Port.getControlEnabled Port.setControlEnabled

instance ToolTip Popup where
  tooltip = newStdAttr phandle Port.getControlTip Port.setControlTip

instance Control Popup where
  pack = stdPack pparent phandle Port.getPopUpRequestSize

instance Commanding Popup where
  command = newStdEvent phandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler

{--------------------------------------------------------------------
  ListBox
--------------------------------------------------------------------}
-- | A list box. Allthough it is 'Commanding', the default
-- command handler automatically calls a handler associated with a
-- selected item.
data ListBox = ListBox { lbhandle  :: !WindowHandle
                       , lbparent  :: !WindowHandle
                       , lbitems   :: Var [(String,IO ())]
                       }

-- | Create a list box.
listBox :: Container w => Bool -> [Prop ListBox] -> w -> IO ListBox
listBox multi props w = do
	lb <- do
			hlist <- Port.createListBox (hwindow w) multi
			lbitems <- newVar []
			return (ListBox hlist (hwindow w) lbitems)
	set lb [on command =: listBoxCommand lb]
	set lb props
	return lb

-- default command handler
listBoxCommand :: ListBox -> IO ()
listBoxCommand lb = do
  i  <- Port.getListBoxCurrentItem (lbhandle lb)
  xs <- getVar (lbitems lb)
  when (i>=0 && i < length xs) (snd (xs!!i))  -- invoke appropiate handler

instance Countable ListBox where
  count = readAttr "count" (fmap length . getVar . lbitems)

instance CommandItems ListBox where
  items
    = newAttr (\w -> getVar (lbitems w))
              (\w xs -> do Port.removeAllListBoxItems (lbhandle w)
                           mapM_ (Port.appendListBoxItem (lbhandle w) . fst) xs
                           setVar (lbitems w) xs)
                           
  appendItem lb item@(title,action) = do
    items <- takeVar (lbitems lb)
    Port.appendListBoxItem (lbhandle lb) title
    putVar (lbitems lb) (items++[item])
  
  insertItem lb pos item@(title,action) = do
    items <- takeVar (lbitems lb)
    Port.insertListBoxItem (lbhandle lb) pos title
    putVar (lbitems lb) (insertAt pos item items)
  
  removeItem lb pos = do
    items <- takeVar (lbitems lb)
    Port.removeListBoxItem (lbhandle lb) pos
    putVar (lbitems lb) (removeAt pos items)     
  
  removeAllItems lb = do
    items <- takeVar (lbitems lb)
    Port.removeAllListBoxItems (lbhandle lb)
    putVar (lbitems lb) []


instance SingleSelect ListBox where
  selected = newStdAttr lbhandle Port.getListBoxSingleSelection Port.setListBoxSingleSelection

instance MultiSelect ListBox where
  selection = newAttr getter setter
    where
      getter w = do
        items <- getVar (lbitems w)
        filterM (\i -> Port.getListBoxItemSelectState (lbhandle w) i) [0..length items-1]

      setter w xs = do
        items <- getVar (lbitems w)
        mapM_ (\x -> Port.setListBoxItemSelectState (lbhandle w) x (elem x xs)) [0..length items-1]

instance Dimensions ListBox where
  frame = newStdAttr lbhandle Port.getControlFrame Port.moveResizeControl

instance Able ListBox where
  enabled = newStdAttr lbhandle Port.getControlEnabled Port.setControlEnabled

instance ToolTip ListBox where
  tooltip = newStdAttr lbhandle Port.getControlTip Port.setControlTip

instance Control ListBox where
  pack = stdPack lbparent lbhandle Port.getListBoxRequestSize

instance Commanding ListBox where
  command = newStdEvent lbhandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler

--------------------------------------------------------------------
--  Check box
--------------------------------------------------------------------
-- | A single check control.
data CheckBox = CheckBox
	{ chandle :: !WindowHandle
	, cparent :: !WindowHandle
	}

-- | Create a check control with a certain label.
checkBox :: Container w => [Prop CheckBox] -> w -> IO CheckBox
checkBox props w
  = do c <- do hcheck <- Port.createCheckBox (hwindow w)
               return (CheckBox hcheck (hwindow w))
       set c props
       return c

instance Titled CheckBox where
  title = newStdAttr chandle Port.getCheckBoxText  Port.setCheckBoxText

instance Checked CheckBox where
  checked = newAttr (\w   -> Port.getCheckBoxSelectState (chandle w))
                    (\w b -> do Port.setCheckBoxSelectState (chandle w) b
                                cmd <- get w (on command)
                                cmd)
   
instance Commanding CheckBox where
  command = newStdEvent chandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler
  
instance Able CheckBox where
  enabled = newStdAttr chandle Port.getControlEnabled Port.setControlEnabled
  
instance ToolTip CheckBox where
  tooltip = newStdAttr chandle Port.getControlTip Port.setControlTip

instance Control CheckBox where
  pack = stdPack cparent chandle Port.getCheckBoxRequestSize

--------------------------------------------------------------------
-- Radio box
--------------------------------------------------------------------
-- | A single radio control.
data RadioBox = RadioBox
	{ rhandle :: !WindowHandle
	, rparent :: !WindowHandle
	}

-- | Create a radio control.
radioBox :: Container w => [Prop RadioBox] -> w -> IO RadioBox
radioBox props w
  = do r <- do hradio <- Port.createRadioBox (hwindow w)
               return (RadioBox hradio (hwindow w))
       set r props
       return r

instance Titled RadioBox where
  title = newStdAttr rhandle Port.getRadioBoxText Port.setRadioBoxText

instance Checked RadioBox where
  checked = newStdAttr rhandle Port.getRadioBoxSelectState Port.setRadioBoxSelectState
   
instance Commanding RadioBox where
  command = newStdEvent rhandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler

instance Able RadioBox where
  enabled = newStdAttr rhandle Port.getControlEnabled Port.setControlEnabled
  
instance ToolTip RadioBox where
  tooltip = newStdAttr rhandle Port.getControlTip Port.setControlTip

instance Control RadioBox where
  pack = stdPack rparent rhandle Port.getRadioBoxRequestSize


-- | Connect a list of radio controls such that only one of them is selected
-- at any time.
setRadioBoxGroup :: [RadioBox] -> IO ()
setRadioBoxGroup items = Port.setRadioBoxGroup (map rhandle items)

--------------------------------------------------------------------
-- Slider
--------------------------------------------------------------------

-- | A slider control.
data Slider = Slider{ shandle :: !WindowHandle
                    , sparent :: !WindowHandle
                    }
                      
-- | Create a horizontal slider control.
hslider :: Container w => [Prop Slider] -> w -> IO Slider
hslider props w
  = do r <- do shandle <- Port.createHorzSlider (hwindow w)
               return (Slider shandle (hwindow w))
       set r props
       return r
       
-- | Create a vertical slider control.
vslider :: Container w => [Prop Slider] -> w -> IO Slider
vslider props w
  = do r <- do shandle <- Port.createVertSlider (hwindow w)
               return (Slider shandle (hwindow w))
       set r props
       return r

instance RangedSelect Slider where
   range = newAttr (\w   -> Port.getSliderRange (shandle w))
                   (\w (min,max) -> Port.setSliderRange (shandle w) min max)

   selectedPos = newStdAttr shandle Port.getSliderPosition Port.setSliderPosition

instance Commanding Slider where
  command = newStdEvent shandle Port.getControlCommandHandler Port.setControlCommandHandler Port.setControlCommandDefHandler

instance Control Slider where
  pack = stdPack sparent shandle Port.getSliderRequestSize


--------------------------------------------------------------------
-- ProgressBar
--------------------------------------------------------------------

-- | A progress bar.
data ProgressBar = ProgressBar
                    { pbhandle :: !WindowHandle
                    , pbparent :: !WindowHandle
                    , pbrange  :: Var (Int,Int)
                    }
                      
-- | Create a horizontal progress bar.
-- The boolean parameter specify whether the bar shows continuous or discrete values.
hProgressBar :: Container w => Bool -> [Prop ProgressBar] -> w -> IO ProgressBar
hProgressBar smooth props w
  = do r <- do pbhandle <- Port.createHorzProgressBar (hwindow w) smooth
               pbrange  <- newVar (0,100)
               Port.setProgressBarFraction pbhandle 0 100 0
               return (ProgressBar pbhandle (hwindow w) pbrange)
       set r props
       return r
       
-- | Create a vertical progress bar.
-- The boolean parameter specify whether the bar shows continuous or discrete values.
vProgressBar :: Container w => Bool -> [Prop ProgressBar] -> w -> IO ProgressBar
vProgressBar smooth props w
  = do r <- do pbhandle <- Port.createVertProgressBar (hwindow w) smooth
               pbrange  <- newVar (0,100)
               Port.setProgressBarFraction pbhandle 0 100 0
               return (ProgressBar pbhandle (hwindow w) pbrange)
       set r props
       return r

instance RangedSelect ProgressBar where
   range = newAttr (\w   -> getVar (pbrange w))
                   (\w r@(min,max) -> do
                       pos <- Port.getProgressBarFraction (pbhandle w) min max
                       Port.setProgressBarFraction (pbhandle w) min max pos
                       setVar (pbrange w) r)

   selectedPos 
         = newAttr (\w -> do
                        (min,max) <- getVar (pbrange w)
                        Port.getProgressBarFraction (pbhandle w) min max)
                   (\w pos -> do
                        (min,max) <- getVar (pbrange w)
                        Port.setProgressBarFraction (pbhandle w) min max pos)

instance Control ProgressBar where
  pack = stdPack pbparent pbhandle Port.getProgressBarRequestSize


--------------------------------------------------------------------
-- CompoundControl
--------------------------------------------------------------------
-- | A compound control.
data CompoundControl = CompoundControl
	{ cchandle     :: !WindowHandle
	, ccparent     :: !WindowHandle
	, vdomain      :: Var Size
	, vautosize    :: Var Bool
	, vpen         :: Var Pen
	, vbufferMode  :: Var BufferMode
	, vpaint       :: Var PaintFunction
	, vlayout      :: Var Layout                     
	}

-- | Create a compound control
compoundControl :: Container w => [Prop CompoundControl] -> w -> IO CompoundControl
compoundControl props w
  = do c <- do cchandle   <- Port.createCompoundControl (hwindow w)
               vpaint     <- newVar (\_ _ _ -> return ())
               vautosize  <- newVar True
               vlayout    <- newVar empty
               vdomain    <- newVar (sz 0 0)
               vpen       <- newVar windowPen
               vbufferMode<- newVar UnBuffered
               return (CompoundControl cchandle (hwindow w) vdomain vautosize vpen vbufferMode vpaint vlayout)
       recolorCompound c
       set c [on relayout =:: relayoutCompound]
       -- just by setting a dummy paint function, we will at least intialize the canvas properly on a repaint
       set c [on paint =: (\_ _ _ -> return ())]
       set c props
       return c

relayoutCompound :: CompoundControl -> IO ()
relayoutCompound c
  = do frame  <- Port.getControlFrame (cchandle c)
       domain <- get c domain
       lay    <- getVar (vlayout c)
       needed <- getLayoutSize lay
       let d1 = maxSize domain needed
           d2 = maxSize d1 (rectSize frame)
       Port.setWindowDomainSize (cchandle c) d1
       layoutInRect (rectOfSize d2) lay
       return ()

recolorCompound :: CompoundControl -> IO ()
recolorCompound c
  = do col   <- get c color
       bgcol <- get c bgcolor
       hat   <- get c hatch
       Port.setWindowColor (cchandle c) col bgcol hat
       repaint c
       relayoutCompound c

instance Scrollable CompoundControl where
  scroll = newStdEvent cchandle Port.getWindowScrollHandler Port.setWindowScrollHandler Port.setWindowScrollDefHandler
  domain = newAttr (\c   -> getVar (vdomain c))
                   (\c x -> setVar (vdomain c) x >> relayoutCompound c)
  origin = newStdAttr cchandle Port.getWindowScrollOrigin Port.setWindowScrollOrigin

instance Dimensions CompoundControl where
  frame  = newStdAttr cchandle Port.getControlFrame Port.moveResizeControl

instance Drawn CompoundControl where
  pen = newAttr (getVar . vpen) (\w pen -> setVar (vpen w) pen >> recolorCompound w)
  
  bufferMode = newStdAttr vbufferMode getVar setVar
  
instance HasFont CompoundControl where
  font = mapAttr penFont (\pen c -> pen{penFont=c}) pen

instance Reactive CompoundControl where
  mouse       = newStdEvent cchandle Port.getWindowMouseHandler       Port.setWindowMouseHandler       Port.setWindowMouseDefHandler
  keyboard    = newStdEvent cchandle Port.getWindowKeyboardHandler    Port.setWindowKeyboardHandler    Port.setWindowKeyboardDefHandler
  contextMenu = newStdEvent cchandle Port.getWindowContextMenuHandler Port.setWindowContextMenuHandler Port.setWindowContextMenuDefHandler

instance Paint CompoundControl where
  repaint c = Port.invalidateWindow (cchandle c)
  paint     = newEvent (getVar . vpaint)
                          (\w h -> Port.setWindowPaintHandler    (cchandle w) (wndpaint w h) >> setVar (vpaint w) h)
                          (\w   -> Port.setWindowPaintDefHandler (cchandle w) >> setVar (vpaint w) (\_ _ _ -> return ()))
            where
              wndpaint c paintfun hcanvas updArea
                = do pen   <- get c pen
                     bmode <- get c bufferMode
                     withCanvas bmode pen hcanvas $ \can -> paintfun can updArea []
  paintIn c bmode f = do 
     pen   <- get c pen
     Port.drawInWindow (cchandle c) (\hcanvas -> withCanvas bmode pen hcanvas f)

instance Able CompoundControl where
  enabled = newStdAttr cchandle Port.getControlEnabled Port.setControlEnabled

instance Container CompoundControl where
  layout 
    = writeAttr "layout" (\w c -> do
       let new_lay = pack c
       old_lay  <- getVar (vlayout w)
       domain   <- get w domain
       needed   <- getLayoutSize new_lay
       let d = maxSize domain needed
       Port.setWindowDomainSize (cchandle w) d
       frame    <- Port.getControlFrame (cchandle w)
       updateControlsVisibility old_lay new_lay
       layoutInRect (rectOfSize (maxSize d (rectSize frame))) new_lay
       setVar (vlayout w) new_lay)
  autosize   = readAttr "autosize"   (\c -> return False)
  layoutSize = readAttr "layoutSize" (\c -> getVar (vlayout c) >>= getLayoutSize)
  relayout   = newStdEvent cchandle Port.getContainerReLayoutHandler Port.setContainerReLayoutHandler Port.setContainerReLayoutDefHandler
  hwindow c  = cchandle c

instance Control CompoundControl where
  pack = stdPack ccparent cchandle Port.getCompoundControlRequestSize


--------------------------------------------------------------------
-- GroupBox
--------------------------------------------------------------------
-- | A group box.
data GroupBox = GroupBox
	{ gbhandle     :: !WindowHandle
	, gbparent     :: !WindowHandle
	, gblayout     :: Var Layout
	}

-- | Create a group box
groupBox :: Container w => [Prop GroupBox] -> w -> IO GroupBox
groupBox props w
  = do c <- do gbhandle   <- Port.createGroupBox (hwindow w)
               gblayout   <- newVar empty
               return (GroupBox gbhandle (hwindow w) gblayout)
       set c [on relayout =:: relayoutGroupBox]
       set c props
       return c

relayoutGroupBox :: GroupBox -> IO ()
relayoutGroupBox c
  = do (Rect fl ft fr fb) <- Port.getControlFrame (gbhandle c)
       lay  <- getVar (gblayout c)
       (l,t,r,b) <- Port.getGroupBoxBordersSize (gbhandle c)
       layoutInRect (Rect l t (fr-fl-r) (fb-ft-b)) lay
       return ()

instance Container GroupBox where
  layout     = writeAttr "layout"  (\w c -> do
                     let new_lay = pack c
                     old_lay <- getVar (gblayout w)
                     updateControlsVisibility old_lay new_lay
                     setVar (gblayout w) new_lay
                     Port.relayoutContainer (gbhandle w))
  autosize   = readAttr "autosize"   (\c -> return True)
  layoutSize = readAttr "layoutSize" (\c -> getVar (gblayout c) >>= getLayoutSize)
  relayout   = newStdEvent gbhandle Port.getContainerReLayoutHandler Port.setContainerReLayoutHandler Port.setContainerReLayoutDefHandler
  hwindow c  = gbhandle c

instance Titled GroupBox where
  title  = newStdAttr gbhandle Port.getGroupBoxText Port.setGroupBoxText

instance Control GroupBox where
  pack c = stdPack gbparent gbhandle getGroupBoxRequestSize c
  	where
		getGroupBoxRequestSize hwnd = do
			(l,t,r,b) <- Port.getGroupBoxBordersSize hwnd
			Size w h <- getVar (gblayout c) >>= getLayoutSize
			return (Size (l+w+r) (t+h+b))



--------------------------------------------------------------------
-- Notebook
--------------------------------------------------------------------
-- | A notebook control.
data Notebook = Notebook
	{ nbhandle :: !WindowHandle
	, nbparent :: !WindowHandle
	, nbpages  :: !(Var [NotebookPage])
	}

-- | Create a notebook control.
notebook :: Container w => [Prop Notebook] -> w -> IO Notebook
notebook props w
  = do nb <- do hbook <- Port.createNotebook (hwindow w)
                refPages <- newVar []
                return (Notebook hbook (hwindow w) refPages)
       set nb props
       return nb

instance Control Notebook where
  pack w = stdPack nbparent nbhandle getNotebookRequestSize w
  		where
  			getNotebookRequestSize hwnd = do
  				outSize <- Port.getNotebookRequestSize hwnd
  				pages <- getVar (nbpages w)
  				pageSizes <- mapM (\page -> get page layoutSize) pages
  				let inSize = foldr maxSize (Size 10 10) pageSizes
  				return (addSize inSize outSize)

labelsPosition :: Attr Notebook PositionType
labelsPosition = newStdAttr nbhandle Port.getNotebookLabelsPosition  Port.setNotebookLabelsPosition

instance Countable Notebook where
  count = readAttr "count" (Port.getNotebookPageCount . nbhandle)

instance SingleSelect Notebook where
  selected = newStdAttr nbhandle Port.getNotebookSelection Port.setNotebookSelection

-- | The selected page in the notebook
selectedPage :: Attr Notebook NotebookPage
selectedPage = newAttr getPage setPage
	where
		getPage w = do
			index <- Port.getNotebookSelection (nbhandle w)
			pages <- getVar (nbpages w)
			return (pages !! index)

		setPage w p = do
			index <- Port.getNotebookPagePos (pghandle p)
			Port.setNotebookSelection (nbhandle w)index

-- | A notebook page.
data NotebookPage = NotebookPage
	{ pghandle :: !WindowHandle
	, pgparent :: !WindowHandle
	, pglayout :: !(Var Layout)
	}

-- | Create a new page at specified position in the notebook.
notebookPageAt :: Maybe Int -> [Prop NotebookPage] -> Notebook -> IO NotebookPage
notebookPageAt mb_pos props (Notebook hNotebook _ refPages)
  = do pg <- do hpage   <- Port.insertNotebookPage hNotebook mb_pos
                vlayout <- newVar empty
                return (NotebookPage hpage hNotebook vlayout)
       updateVar refPages ((:) pg)
       set pg [on relayout =: relayoutNotebookPage pg]
       set pg props
       return pg

-- | Appends a new page in the notebook.
notebookPage :: [Prop NotebookPage] -> Notebook -> IO NotebookPage
notebookPage = notebookPageAt Nothing

relayoutNotebookPage :: NotebookPage -> IO ()
relayoutNotebookPage c
  = do size <- Port.getNotebookPageSize (pghandle c)
       lay  <- getVar (pglayout c)
       layoutInRect (rectOfSize size) lay
       return ()

instance Container NotebookPage where
  layout     = writeAttr "layout"  (\w c -> do
                     let new_lay = pack c
                     old_lay <- getVar (pglayout w)
                     updateControlsVisibility old_lay new_lay                     
                     setVar (pglayout w) new_lay
                     Port.relayoutContainer (pghandle w))
  autosize   = readAttr "autosize"   (\c -> return True)
  layoutSize = readAttr "layoutSize" (\c -> getVar (pglayout c) >>= getLayoutSize)
  relayout   = newStdEvent pghandle Port.getContainerReLayoutHandler Port.setContainerReLayoutHandler Port.setContainerReLayoutDefHandler
  hwindow c  = pghandle c

instance Titled NotebookPage where
  title  = newStdAttr pghandle Port.getNotebookPageTitle Port.setNotebookPageTitle

instance Positioned NotebookPage where
  pos = readAttr "pos" (Port.getNotebookPagePos . pghandle)

instance Activity NotebookPage where
  activate  = newStdEvent pghandle Port.getWindowActivateHandler   Port.setWindowActivateHandler   Port.setWindowActivateDefHandler
  deactivate= newStdEvent pghandle Port.getWindowDeactivateHandler Port.setWindowDeactivateHandler Port.setWindowDeactivateDefHandler
  isactive = newAttr getIsActive setIsActive
  	where
  		getIsActive p = do
  			index <- Port.getNotebookPagePos   (pghandle p)
  			sel   <- Port.getNotebookSelection (pgparent p)
  			return (index == sel)

  		setIsActive p _ = do
			index <- Port.getNotebookPagePos (pghandle p)
			Port.setNotebookSelection (pgparent p)index

instance Deadly NotebookPage where
  destroyWidget w = Port.destroyNotebookPage (pghandle w)
  destroy         = newStdEvent pghandle Port.getWindowDestroyHandler Port.setWindowDestroyHandler Port.setWindowDestroyDefHandler

instance HasIcon NotebookPage where
  icon  = newStdAttr pghandle Port.getNotebookPageBitmap Port.setNotebookPageBitmap
