-----------------------------------------------------------------------------------------
{-| Module      :  Menu
    Copyright   :  (c) Krasimir Angelov 2003
    License     :  BSD-style

    Maintainer  :  ka2_mail@yahoo.com
    Stability   :  provisional
    Portability :  portable

    The module contains all utilitites required for creation and management of menus.
    A menu is a list of items that specify options or groups of options (a submenu). 
    Clicking a menu item opens a submenu or causes the application to carry out a command. 
    A menu is arranged in a hierarchy. At the top level of the hierarchy is the menu bar; 
    which contains a list of menus, which in turn can contain submenus. 
-}
-----------------------------------------------------------------------------------------
module Graphics.UI.GIO.Menu
            ( -- * Menus
              Menu, mainMenu, menuAt, menu
            , popupMenu, trackPopupMenu
              -- * Menu items
              -- ** Command items
            , MenuItem, menuitemAt, menuitem
	      -- ** Checked items
            , MenuCheck, menucheckAt, menucheck
              -- ** Radio items
            , MenuRadioItem, menuRadioItemAt, menuRadioItem
            , setMenuRadioGroup
              -- ** Menu separator
            , MenuLine, menulineAt, menuline

              -- internal
            , hmenu
            ) where

import System.IO.Unsafe( unsafePerformIO )
import qualified Graphics.UI.Port as Lib
import Graphics.UI.GIO.Types
import Graphics.UI.GIO.Attributes
import Graphics.UI.GIO.Layout
import Graphics.UI.GIO.Events
import Graphics.UI.GIO.Window

{--------------------------------------------------------------------
  Menus
--------------------------------------------------------------------}
-- | A menu is a widget that contains menu items or other (sub) menus.
-- The top level menu is always a 'mainMenu'. 
newtype Menu = Menu MenuHandle
hmenu (Menu h) = h

-- | The main application menu (the menu bar)
mainMenu :: Menu
mainMenu = Menu Lib.mainMenu

-- | The 'menuAt' function creates and inserts an item with submenu in the parent menu. 
-- The created submenu is initially empty. You can populate it with items by using the 
-- 'menuitemAt', 'menucheckAt', 'menuRadioItemAt' or 'menulineAt' functions.
-- Using the 'menuAt' you can create nested submenus.
menuAt :: Maybe Int                               -- ^ The position where to place the submenu
                                                  -- or Nothing if you want to append the item.
       -> [Prop Menu]                             -- ^ The setup of the menu attributes
       -> Menu                                    -- ^ The parent menu
       -> IO Menu                                 -- ^ The created sub menu
menuAt pos props parent
  = do m     <- do hmenu <- Lib.insertMenu (hmenu parent) pos
                   return (Menu hmenu)
       set m props
       return m

-- | The function is the same as the 'menuAt' function but always appends the item. The 'menu' function
-- is semantically equal to @menuAt Nothing@
menu :: [Prop Menu] -> Menu -> IO Menu
menu = menuAt Nothing

-- | The 'popupMenu' function creates a popup menu. The menu is initially empty.
-- You can populate it with items by using the 'menuitemAt', 'menucheckAt', 
-- 'menuRadioItemAt' or 'menulineAt' functions. Using the 'menuAt' you can create nested submenus.
popupMenu :: [Prop Menu] -> IO Menu
popupMenu props
  = do m     <- do hmenu <- Lib.createPopupMenu
                   return (Menu hmenu)
       set m props
       return m

-- | The 'trackPopupMenu' function displays a popup menu at the specified location in the
-- window and tracks the selection of items on the menu.
trackPopupMenu :: Menu                      	  -- ^ The popup menu
               -> Window                          -- ^ The window in which to popup the menu
               -> Point                           -- ^ The location inside the window at which to display the menu
               -> IO ()
trackPopupMenu menu window pos = Lib.trackPopupMenu (hmenu menu) (hwindow window) pos

instance Able Menu where
  enabled   = newStdAttr hmenu Lib.getMenuItemEnabled Lib.setMenuItemEnabled

instance Titled Menu where
  title = newStdAttr hmenu Lib.getMenuLabel Lib.setMenuLabel

instance Countable Menu where
  count = readAttr "count" (Lib.getMenuItemCount . hmenu)

instance Positioned Menu where
  pos = readAttr "pos" (Lib.getMenuItemPos . hmenu)
  
instance Deadly Menu where
  destroyWidget m = Lib.destroyMenu (hmenu m)
  destroy         = newStdEvent hmenu Lib.getMenuDestroyHandler Lib.setMenuDestroyHandler Lib.setMenuDestroyDefHandler

--------------------------------------------------------------------
--  Menu items
--------------------------------------------------------------------
-- | Menu items are labeled entries in a menu.
newtype MenuItem = MenuItem MenuHandle
hitem (MenuItem h) = h

-- | Create a menu item and insert it at specified position.
menuitemAt :: Maybe Int -> [Prop MenuItem] -> Menu -> IO MenuItem
menuitemAt pos props menu
  = do mitem <- do hitem   <- Lib.insertMenuItem (hmenu menu) pos
                   return (MenuItem hitem)
       set mitem props
       return mitem

-- | Create a menu item and appends it to parent menu.
menuitem :: [Prop MenuItem] -> Menu -> IO MenuItem
menuitem = menuitemAt Nothing 

instance Able MenuItem where
  enabled   = newStdAttr hitem Lib.getMenuItemEnabled Lib.setMenuItemEnabled

instance Commanding MenuItem where
  command = newStdEvent hitem Lib.getMenuCommandHandler Lib.setMenuCommandHandler Lib.setMenuCommandDefHandler
    
instance DynamicUpdate MenuItem where
  update  = newStdEvent hitem Lib.getMenuUpdateHandler  Lib.setMenuUpdateHandler  Lib.setMenuUpdateDefHandler

instance Titled MenuItem where
  title = newStdAttr hitem Lib.getMenuLabel Lib.setMenuLabel

instance Accelerated MenuItem where
  accel = newStdAttr hitem Lib.getMenuItemAccel  Lib.setMenuItemAccel

instance HasIcon MenuItem where
  icon  = newStdAttr hitem Lib.getMenuItemBitmap Lib.setMenuItemBitmap

instance Positioned MenuItem where
  pos = readAttr "pos" (Lib.getMenuItemPos . hitem)

instance Deadly MenuItem where
  destroyWidget m = Lib.destroyMenu (hitem m)
  destroy         = newStdEvent hitem Lib.getMenuDestroyHandler Lib.setMenuDestroyHandler Lib.setMenuDestroyDefHandler
  
--------------------------------------------------------------------
--  Menu radio items
--------------------------------------------------------------------
  
-- | Radio menu items are labeled entries in a menu with bookmark.
-- Sometimes, a group of menu items corresponds to a set of mutually exclusive options.
-- In this case, you can indicate the selected option by using a selected radio menu item.
-- To check a menu item use the 'checked' attribute.
newtype MenuRadioItem = MenuRadioItem MenuHandle
hradioitem (MenuRadioItem h) = h

-- | Create a radio menu item and insert it at specified position.
menuRadioItemAt :: Maybe Int                      -- ^ The position where to place the item
                                                  -- or Nothing if you want to append it. 
       -> [Prop MenuRadioItem]                    -- ^ The setup of the radio item attributes
       -> Menu                                    -- ^ The parent menu
       -> IO MenuRadioItem                        -- ^ The created radio item
menuRadioItemAt pos props menu
  = do mradioitem <- do hradioitem   <- Lib.insertMenuRadioItem (hmenu menu) pos
                        return (MenuRadioItem hradioitem)
       set mradioitem props
       return mradioitem

-- | The function is the same as the 'menuRadioItemAt' function but always appends the item. The 'menuRadioItem' 
-- function is semantically equal to @menuRadioItemAt Nothing@
menuRadioItem :: [Prop MenuRadioItem] -> Menu -> IO MenuRadioItem
menuRadioItem = menuRadioItemAt Nothing 

instance Able MenuRadioItem where
  enabled   = newStdAttr hradioitem Lib.getMenuItemEnabled Lib.setMenuItemEnabled

instance Commanding MenuRadioItem where
  command = newStdEvent hradioitem Lib.getMenuCommandHandler Lib.setMenuCommandHandler Lib.setMenuCommandDefHandler
    
instance DynamicUpdate MenuRadioItem where
  update  = newStdEvent hradioitem Lib.getMenuUpdateHandler  Lib.setMenuUpdateHandler  Lib.setMenuUpdateDefHandler

instance Positioned MenuRadioItem where
  pos = readAttr "pos" (Lib.getMenuItemPos . hradioitem)
  
instance Checked MenuRadioItem where
  checked = newStdAttr hradioitem Lib.getMenuItemChecked Lib.setMenuItemChecked

instance Deadly MenuRadioItem where
  destroyWidget m = Lib.destroyMenu (hradioitem m)
  destroy         = newStdEvent hradioitem Lib.getMenuDestroyHandler Lib.setMenuDestroyHandler Lib.setMenuDestroyDefHandler
  
instance Titled MenuRadioItem where
  title = newStdAttr hradioitem Lib.getMenuLabel Lib.setMenuLabel

instance Accelerated MenuRadioItem where
  accel = newStdAttr hradioitem Lib.getMenuItemAccel Lib.setMenuItemAccel

-- | The 'setMenuRadioGroup' function specifies a set of mutually exclusive options.
setMenuRadioGroup :: [MenuRadioItem] -> IO ()
setMenuRadioGroup items = Lib.setMenuRadioGroup (map hradioitem items)
  
--------------------------------------------------------------------
--  Checked menu items
--------------------------------------------------------------------
-- | Checked menu items are labeled entries in a menu with check mark.
-- Applications typically check or clear a menu item to indicate whether 
-- an option is in effect.
newtype MenuCheck = MenuCheck MenuHandle
hcheck (MenuCheck h) = h

-- | Create a check menu item and insert it at specified position.
menucheckAt :: Maybe Int                      -- ^ The position where to place the item
                                              -- or Nothing if you want to append it.
            -> [Prop MenuCheck]               -- ^ The setup of the item attributes
            -> Menu                           -- ^ The parent menu
            -> IO MenuCheck                   -- ^ The created checked item
menucheckAt pos props menu
  = do mcheck <- do mcheck <- Lib.insertMenuCheckItem (hmenu menu) pos
                    return (MenuCheck mcheck)
       set mcheck props
       return mcheck

-- | The function is the same as the 'menucheckAt' function but always appends the item. The 'menucheck'
-- function is semantically equal to @menucheckAt Nothing@
menucheck :: [Prop MenuCheck] -> Menu -> IO MenuCheck
menucheck = menucheckAt Nothing 

instance Able MenuCheck where
  enabled   = newStdAttr hcheck Lib.getMenuItemEnabled Lib.setMenuItemEnabled

instance Commanding MenuCheck where
  command = newStdEvent hcheck Lib.getMenuCommandHandler Lib.setMenuCommandHandler Lib.setMenuCommandDefHandler
    
instance DynamicUpdate MenuCheck where
  update  = newStdEvent hcheck Lib.getMenuUpdateHandler  Lib.setMenuUpdateHandler  Lib.setMenuUpdateDefHandler
  
instance Titled MenuCheck where
  title = newStdAttr hcheck Lib.getMenuLabel Lib.setMenuLabel

instance Accelerated MenuCheck where
  accel = newStdAttr hcheck Lib.getMenuItemAccel Lib.setMenuItemAccel

instance Positioned MenuCheck where
  pos = readAttr "pos" (Lib.getMenuItemPos . hcheck)
  
instance Checked MenuCheck where
  checked = newStdAttr hcheck Lib.getMenuItemChecked Lib.setMenuItemChecked

instance Deadly MenuCheck where
  destroyWidget m = Lib.destroyMenu (hcheck m)
  destroy         = newStdEvent hcheck Lib.getMenuDestroyHandler Lib.setMenuDestroyHandler Lib.setMenuDestroyDefHandler

--------------------------------------------------------------------
--  Menu separator
--------------------------------------------------------------------

-- | Menu separator item
newtype MenuLine = MenuLine MenuHandle
hline (MenuLine h) = h

-- | Insert a menu seperator line at specified position.
menulineAt :: Maybe Int -> Menu -> IO MenuLine
menulineAt pos menu
  = do hitem  <- Lib.insertMenuSeparatorItem (hmenu menu) pos
       return (MenuLine hitem)

-- | Append a menu seperator line
menuline :: Menu -> IO MenuLine
menuline = menulineAt Nothing

instance Positioned MenuLine where
  pos = readAttr "pos" (Lib.getMenuItemPos . hline)

instance Deadly MenuLine where
  destroyWidget m = Lib.destroyMenu (hline m)
  destroy         = newStdEvent hline Lib.getMenuDestroyHandler Lib.setMenuDestroyHandler Lib.setMenuDestroyDefHandler
